#!/bin/bash
#
# Script for caching SMV results
#
# Author: Nicolas Blanc
#

if [ ! $# == 1 ]; then
	echo "One parameter expected but got $#."
	exit 1
fi

if [ ! -e $1 ]; then
	echo "File $1 does not exit."
	exit 1
fi

CACHE_DIR=.SMV_CACHE

if [ ! -d $CACHE_DIR ]; then
	echo "Creating cache directory: $CACHE_DIR"
	mkdir $CACHE_DIR
else
FILES=`ls -1 $CACHE_DIR/*.smv*`
fi

BASE_NAME=`echo $1 |  sed "s/\.smv//"`
echo "BASE_NAME $BASE_NAME"



for FILE in $FILES; do
	diff -q $FILE $1 > /dev/null
	if [ $? == 0 ]; then
		echo "Cache hit."
		BASE_IN_CACHE=`echo "$FILE" | sed s/\.smv*\.*//`
		echo "BASE_IN_CACHE $BASE_IN_CACHE"
		VERSION_IN_CACHE=`echo "$FILE" | sed s/.*\.smv//`
		echo "VERSION_IN_CACHE $VERSION_IN_CACHE"
		cp $BASE_IN_CACHE.out$VERSION_IN_CACHE     $BASE_NAME.out
		cp $BASE_IN_CACHE.stats$VERSION_NAME       $BASE_NAME.stats
		cp $BASE_IN_CACHE.update$VERSION_NAME      $BASE_NAME.update
		cp $BASE_IN_CACHE.warn$VERSION_NAME        $BASE_NAME.warn
		exit 0
	fi
done

smv -sift $1
SMV_EXIT=$?

if [ $SMV_EXIT == 0 ]; then
	cp --backup=t $BASE_NAME.smv $BASE_NAME.out  $BASE_NAME.stats  $BASE_NAME.update $BASE_NAME.warn $CACHE_DIR
fi
exit $SMV_EXIT
