//===========================================
// Function : Asynchronous read write RAM
// Coder    : Deepak Kumar Tala
// Date     : 18-April-2002
//===========================================
#include "systemc.h"

#define DATA_WIDTH        8 
#define ADDR_WIDTH        8 
#define RAM_DEPTH         256

SC_MODULE (ram_dp_ar_aw) {
  sc_in    <sc_uint<ADDR_WIDTH> > address_0;
  sc_in    <bool> cs_0 ;
  sc_in    <bool> we_0 ;
  sc_in    <bool> oe_0  ;
  sc_in    <sc_uint<ADDR_WIDTH> > address_1;
  sc_in    <bool> cs_1 ;
  sc_in    <bool> we_1 ;
  sc_in    <bool> oe_1 ;

  sc_in<sc_uint<DATA_WIDTH> > data_0_in;
  sc_out<sc_uint<DATA_WIDTH> > data_0_out;

  sc_in<sc_uint<DATA_WIDTH> > data_1_in;
  sc_out<sc_uint<DATA_WIDTH> > data_1_out;

  //-----------Internal variables-------------------
  static sc_uint <DATA_WIDTH> mem [RAM_DEPTH];

  //-----------Methods------------------------------
  void  READ_0 ();
  void  READ_1 ();
  void  WRITE_0 ();
  void  WRITE_1 ();

  //-----------Constructor--------------------------
  SC_CTOR(ram_dp_ar_aw) {
    SC_METHOD (READ_0);
      sensitive << address_0 << cs_0 << we_0 << oe_0;
    SC_METHOD (READ_1);
      sensitive << address_1 << cs_1 << we_1 << oe_1;
    SC_METHOD (WRITE_0);
      sensitive << address_0 << cs_0 << we_0;
    SC_METHOD (WRITE_1);
      sensitive << address_1 << cs_1 << we_1;
  }
};

void  ram_dp_ar_aw::READ_0 () {
    if (cs_0.read() && oe_0.read() && !we_0.read()) {
          data_0_out = mem[address_0.read()];
     }
}

void  ram_dp_ar_aw::READ_1 () {
    if (cs_1.read() && oe_1.read() && !we_1.read()) {
          data_1_out = mem[address_1.read()];
     }
}

void  ram_dp_ar_aw::WRITE_0 () {
    if (cs_0.read() && we_0.read()) {
          mem[address_0.read()] = data_0_in.read();
     } 
}

void  ram_dp_ar_aw::WRITE_1 () {
    if (cs_1.read() && we_1.read()) {
          mem[address_1.read()] = data_1_in.read();
     } 
}


SC_MODULE(top)
{
  sc_signal<sc_uint<ADDR_WIDTH> > address_0;
  sc_signal<bool> cs_0 ;
  sc_signal<bool> we_0 ;
  sc_signal<bool> oe_0 ;
  sc_signal<sc_uint<ADDR_WIDTH> > address_1;
  sc_signal<bool> cs_1 ;
  sc_signal<bool> we_1 ;
  sc_signal<bool> oe_1 ;

  sc_signal<sc_uint<DATA_WIDTH> > data_0_out;
  sc_signal<sc_uint<DATA_WIDTH> > data_0_in;

  sc_signal<sc_uint<DATA_WIDTH> > data_1_out;
  sc_signal<sc_uint<DATA_WIDTH> > data_1_in;

  sc_clock clk;

  ram_dp_ar_aw ram;

  SC_CTOR(top):ram("ram")
  {
    ram.address_0(address_0);
    ram.cs_0(cs_0);
    ram.we_0(we_0);
    ram.oe_0(oe_0);
    ram.data_0_in(data_0_in);
    ram.data_0_out(data_0_out);

    ram.address_1(address_1);
    ram.cs_1(cs_1);
    ram.we_1(we_1);
    ram.oe_1(oe_1);
    ram.data_1_in(data_1_in);
    ram.data_1_out(data_1_out);

    SC_CTHREAD(run,clk);
  }

  void run()
  {
    sc_uint<6> ctrl_count = 0x20;

    while(true)
    {
      wait();

      // Port 0
      cs_0 = ctrl_count[5];
      oe_0 = ctrl_count[4];
      we_0 = ctrl_count[3];
      address_0.write(address_0.read() + 32);
      data_0_in.write(data_0_in.read() + 1024);

      // Port 1
      cs_1 = ctrl_count[2];
      oe_1 = ctrl_count[1];
      we_1 = ctrl_count[0];      
      address_1.write(address_1.read() + 128);
      data_1_in.write(data_1_in.read() + 1048576);

      ctrl_count++;
    }
  }
};

int sc_main(int n, char* argv[])
{
  top t("top");
  
  int time;
  std::cin >> time;

  sc_start(time);
}
