/*******************************************************************\
 *
 *
 * Author: Blanc Nicolas
 *
 * 
\*******************************************************************/


#ifndef __SCHEDULE_H
#define __SCHEDULE_E

#include <__signedbv.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <stdlib.h>
#include <set>
#include <map>
#include <list>
#include <unistd.h>
#include <stdio.h>


//#define DEBUG

#ifdef DEBUG
#define IF_DEBUG(x) x
#else
#define IF_DEBUG(x) ;
#endif

__signedbv<32> __scoot_next_schedule( bool runnable_table[],
    bool (**pcond_indep_table)(),
    bool (**pglobal_indep_table)(), __signedbv<32> n )
{

  std::set<int> runnable;
  static std::set<int> sleeps;
  static std::set<int> independents;
  static bool new_evaluation_phase = true;

#ifndef RANDOM_SEARCH
  static std::ofstream pout;
  static std::ifstream pin;
#endif

  // initialization code
  static bool initialized = false;
  if(!initialized)
  {
    initialized = true;
#ifdef RANDOM_SEARCH
    //		unsigned seed;	
    //		std::cout << "Random seed? ";
    //		std::cin  >> seed;
#else
    pout.open("./scoot_sim.pipe.out", std::ofstream::out);
    pin.open("./scoot_sim.pipe.in", std::ifstream::in);
#endif

  }

  if(new_evaluation_phase == true)
  {
    for(int pi = 0; pi < n.to_int(); pi++)
    {
      independents.insert(pi);
      for(int pj = 0; pj < n.to_int(); pj++)
      {
        if(pi == pj) continue;
        bool (*pfunc)() = pglobal_indep_table[pi*n.to_int()+pj]; 
        bool indep = (*pfunc)();
        if(!indep)
        {
          independents.erase(pi);
          break;
        }
      }
    }
    new_evaluation_phase = false;
  }

#ifdef DEBUG
    std::cout << "Independent processes: ";
    for(std::set<int>::const_iterator pi_it = independents.begin(); pi_it != independents.end(); pi_it++)
    {
      std::cout << *pi_it << " ";
    }
    std::cout << std::endl;
#endif


  for(int i = 0; i < n.to_int(); i++)
  {
    if(runnable_table[i])
      runnable.insert(i);
  }


  IF_DEBUG(std::cout << "Runnable processes: " << runnable.size() << std::endl);

  if(runnable.empty())
  {
    assert(sleeps.empty());
    new_evaluation_phase = true;
    return __signedbv<32>(-1);
  }

  // Persistent
  std::set<int> persistents = runnable;

#ifdef PERSISTENT_SET
  // Pick up a globally independent process
  for(std::set<int>::const_iterator pi_it = runnable.begin(); pi_it != runnable.end(); pi_it++)
  {
    if(independents.count(*pi_it) != 0 )
    {
      persistents.clear();
      persistents.insert(*pi_it);
      break;
    }
  }
  if(persistents.size() != 1)
  {
    // Check current processes commutativity 
    assert(persistents.size() == runnable.size());
    persistents.clear();

    for(std::set<int>::const_iterator pi_it = runnable.begin(); pi_it != runnable.end(); pi_it++)
    {
      int pi = *pi_it;
       // We assume that all processes are runnable
      for(int pj = 0; pj < n.to_int(); pj++)
      {
        if(pi == pj) continue;

        bool (*pfunc)() = pcond_indep_table[pi*n.to_int()+pj]; 
        bool indep = (*pfunc)();
        if(indep)
        {
          if(persistents.count(pi) == 0)
            persistents.insert(pj);
        }
        else
        {
          persistents.insert(pi);
          persistents.insert(pj);
        }
      }
    }
    // remove non-runnable processes
    for(std::set<int>::iterator it = persistents.begin();
        it != persistents.end();)
    {
      std::set<int>::iterator curr_it = it++;
      int p = * curr_it;
      if(runnable_table[p] == false)
        persistents.erase(curr_it);
    }
    // if only one process is runnable then persistent is empty.
    if(persistents.empty())
    {
       assert(runnable.size() == 1);
       persistents = runnable;
    }
  }
#endif

  IF_DEBUG(std::cout <<"Persistent set: " << persistents.size() << " proc." << std::endl);
  IF_DEBUG(std::cout <<"Sleep set: " << sleeps.size() << " proc." << std::endl);

  std::set<int> awakes;

  for(std::set<int>::iterator it = persistents.begin(); it != persistents.end(); it++)
  {
    if(sleeps.count(*it) == 0)
      awakes.insert(*it);
  }

  if(awakes.empty())
  {
    IF_DEBUG(std::cout << "Set awakes is empty" << std::endl);
    exit(0);
  }
  std::map<int, std::set<int> > next_sleeps;

#ifdef SLEEP_SET
  for(std::set<int>::const_iterator it1 = awakes.begin();
      it1 != awakes.end(); it1++)
  {
    int pi = *it1;
    for(std::set<int>::const_iterator it2 = sleeps.begin();
        it2 != sleeps.end(); it2++)
    {
      int pj = *it2;
      bool (*pfunc)() = pcond_indep_table[pi*n.to_int()+pj]; 
      bool indep = (*pfunc)();
      if(indep) next_sleeps[pi].insert(pj);
      else
        std::cout << "Race detected between " << pi << " and " << pj << "\n";
    }
    IF_DEBUG(std::cout << "Next sleep set of " << pi << ": " << next_sleeps[pi].size() << "\n");
    sleeps.insert(pi);
  }
#endif

#ifdef RANDOM_SEARCH
  // pickup process randomly
  {
    int rnd = random();
    std::cout << "rnd " << rnd <<std::endl;
    int inc = RAND_MAX/awakes.size();
    std::set<int>::const_iterator pit = awakes.begin();
    for(int counter = inc; counter < rnd; counter += inc)
    {
      pit++;
    }
    assert( pit != awakes.end());
    sleeps = next_sleeps[*pit];
    runnable.erase(*pit);

    IF_DEBUG(std::cout << "scheduling " << *pit << std::endl);
    runnable_table[*pit] = false;
    return __signedbv<32>(*pit);
  }
#else
  IF_DEBUG(std::cout << "Simulator sending " << awakes.size() << std::endl);
  pout << awakes.size() << std::endl;
  pout.flush();
  int next;
  IF_DEBUG(std::cout << "Simulator: Waiting for reply\n");
  pin >> next;
  IF_DEBUG(std::cout << "Simulator: recieved " << next << std::endl);

  assert(next >=0 && next < awakes.size());
  std::set<int>::iterator ita = awakes.begin();
  for(int i =0; i < next; i++)ita++;

  IF_DEBUG(std::cout << "scheduling " << *ita << std::endl);

  sleeps = next_sleeps[*ita];

  runnable_table[*ita] = false;
  return __signedbv<32>(*ita);
#endif

}

#endif
